from turbogears import controllers, expose, validate, error_handler
from model import *
from turbogears import identity, redirect, validators as v
from cherrypy import request, response
from turbogears import config
from turbogears import url, flash
from turbogears.widgets import *
# from shutterbug import json
#import logging
from cStringIO import StringIO
import Image as IMG
import random
import os
#import traceback

#log = logging.getLogger("shutterbug.controllers")

class AddAlbumFields(WidgetsList):
	name = TextField(label='Album Name : ')
	desc = TextArea(label='Description : ', rows=5, cols=40)
	id = HiddenField(default='new')

class AddAlbumSchema(v.Schema):
	name = v.UnicodeString(max=256)
	desc = v.UnicodeString(max=10240)
	id = v.Any(v.Int(not_empty=True),v.OneOf(['new']))

add_album_form = TableForm(fields = AddAlbumFields(), 
			 validator = AddAlbumSchema(),
			 action=url('/album/save'))

def albumslist():
	return [ (album.id, album.name) for album in Album.select() ]

class AddPictureFields(WidgetsList):
	name = TextField(label = 'Picture Title : ')
	desc = TextArea(label = 'Picture Description :', rows=5, cols=40)
	image = FileField(label = 'Filename : ')
	album = SingleSelectField(label = 'Album : ', options=albumslist, 
		validator=v.Int(not_empty=True))

class AddPictureSchema(v.Schema):
	name = v.UnicodeString(max=256)
	desc = v.UnicodeString(max=10240)
	album = v.Int(not_empty=True)

add_photo_form = TableForm(fields=AddPictureFields(),
			validator = AddPictureSchema(),
			action=url('/image/save'))


class EditPictureFields(WidgetsList):
	name = TextField(label = 'Picture Title : ')
	desc = TextArea(label = 'Picture Description :', rows=5, cols=40)
	album = SingleSelectField(label = 'Album : ', options=albumslist, 
		validator=v.Int(not_empty=True))
	id = HiddenField()

class EditPictureSchema(v.Schema):
	name = v.UnicodeString(max=256)
	desc = v.UnicodeString(max=10240)
	album = v.Int(not_empty=True)
	id = v.Int(not_empty=True)

edit_photo_form = TableForm(fields=EditPictureFields(),
			validator = EditPictureSchema(),
			action=url('/image/update'))


class AddCommentFields(WidgetsList):
	comment = TextArea(label = '', rows=5, cols=40, validator=v.UnicodeString(max=20480))

def paginate(page, length, perpage=8):
	# XXXjc this is waaaay too complicated..it's 3am when 
	# you're writing this, and it shows. 
	# if the user put in a 0, move to page 1.
	if page <= 0:
		page = 1
		# page is what the user sees, p is the index we use (zero-based)
	p = page-1
	pages = length/perpage
	if length%perpage == 0 and pages != 0:
		pages = pages - 1 
	if p > pages:
		p = pages
	# again, start is used as an index.
	start_idx = p*perpage
	end_idx = start_idx+perpage
	if end_idx > length:
		end_idx = length

	return (p+1, pages+1, start_idx, end_idx)


class ShowAlbum(controllers.Controller):
	@expose('.templates.slideshow')
	@validate(validators=dict(id=v.Int(not_empty=True),shuffle=v.Bool()))
	def slideshow(self, id, shuffle=False, tg_errors=None):
		if tg_errors:
			if 'id' in tg_errors.keys():
				flash('Error: ID is not an integer.')
				raise redirect(request.headers.get('Referer', url('/')))
			else:
				shuffle=False

		album = Album.get(id)
		return dict(album=album, shuffle=shuffle)

	@expose()
	@identity.require(identity.in_group('admin'))
	@validate(add_album_form)
	def save(self, name, desc, id, tg_errors=None):
		if tg_errors:
			#log.debug(tg_errors)
			flash('Error Creating Album: %s: %s' % \
				(tg_errors.keys()[0], tg_errors.values()[0]))
			raise redirect(request.headers.get('Referer', url('/')))

		if id == 'new':
			album = Album(name=name, descr=desc)
			flash("Album created (%d)." % album.id)
		else:
			#log.debug(type(id))
			#log.debug(id)
			album = Album.get(int(id))
			album.name = name
			album.descr = desc
			flash("Album updated.")

		raise redirect(url('/album/%d' % album.id))

	@expose()
	@identity.require(identity.in_group('admin'))
	@validate(validators = dict(id=v.Int(not_empty=True)))
	def delete(self, id, tg_errors=None):
		if tg_errors:
			flash('Error: ID is not an integer.')
			raise redirect(request.headers.get('Referer', url('/')))

		try:
			album = Album.get(id)
		except:
			flash('Error: Album %d does not exist.' %id)
			raise redirect(request.headers.get('Referer', url('/')))

		for photo in album.photos():
			fullname = os.path.join(config.get('photo_dir'), \
								photo.filename)
			if os.path.exists(fullname):
				os.remove(fullname)

			try:
				photo.destroySelf()
			except:
				pass
				#log.debug(traceback.format_exc())
		
		# get rough page number to return to after delete.
		allalbums=list(Album.select(orderBy="-date"))
		idx = allalbums.index(album)
		page = (idx/8)+1
		
		album.destroySelf()
		flash('Album deleted successfully.')

		raise redirect(url('/?page=%d' % page))

	@expose()
	@validate(validators = dict(id=v.Int(not_empty=True)))
	def randomthumb(self, id, tg_errors=None):
		if tg_errors:
			flash('Error: ID is not an integer.')
			raise redirect(url('/image/thumb/0'))

		try:
			album = Album.get(id)
		except:
			flash('Error: Album %d does not exist.' %id)
			raise redirect(url('/image/thumb/0'))

		index = int(random.random()*album.size())
		if album.size():
			return redirect(url('/image/thumb/%d' % \
						album.photos()[index].id))
		else:
			return redirect(url('/image/thumb/0'))

	@expose(template=".templates.album")
	@validate(validators = dict(id=v.Int(not_empty=True), page=v.Int()))
	def default(self, id, page='1', tg_errors=None):
		if tg_errors:
			if tg_errors['id']:
				flash('Error: ID is not an integer.')
				raise redirect(url('/'))
			if tg_errors['page']:
				flash('Page number invalid, setting to 1.')
				page = 1

		try:
			album = Album.get(id)
		except:
			flash('Error: Album %d does not exist.' % id)
			raise redirect(url('/'))

		allphotos = album.photos()
		page, pages, start, end = paginate(page, len(allphotos))
		photos=allphotos[start:end]

		ea_values = dict(name=album.name, desc=album.descr, \
							id=str(album.id))
		ea_text = 'Save'

		addphoto = TableForm(fields=AddPictureFields(),
			action=url('/image/save'))
		ap_values = dict(album = album.id)
		ap_text = 'Add Photo'

		return dict(album=album, photos=photos, \
					page=page, pages=pages,
					editalbum = add_album_form, 
					editalbum_values=ea_values,
					editalbum_text = ea_text,
					addphoto = addphoto,
					addphoto_values = ap_values,
					addphoto_text = ap_text
					)

class Image(controllers.Controller):
	def __init__(self):
		self.photo_dir = config.get('photo_dir')
		return

	@expose()
	@identity.require(identity.in_group('admin'))
	@validate(edit_photo_form)
	def update(self, name, desc, album, id, tg_errors = None):
		if tg_errors:
			flash('Error: %s: %s' % (tg_errors.keys()[0], 
						 tg_errors.valies()[0]))
			return redirect(url('/image/page/%d' % id))
		
		try:
			photo = Photo.get(id)
		except:
			flash('Photo %d does not exist.' % id)
			return redirect(url('/album/%d' % album))

		photo.name = name
		photo.descr = desc
		photo.albumID = album
		flash('Photo updated.')
		return redirect(url('/image/page/%d' % photo.id))

	@expose()
	@identity.require(identity.in_group('admin'))
	@validate(add_photo_form)
	def save(self, name, desc, image, album, tg_errors = None):
		if tg_errors:
			flash('Error: %s: %s' % (tg_errors.keys()[0], 
						 tg_errors.valies()[0]))
			return redirect(url('/album/%d' % album))

		try:
			a = Album.get(album)
		except:
			flash('Error: album: Does not exist.')
			return redirect(url('/'))
	
		try:
			photo = Photo(name=name, descr=desc, filename='', \
				width=0, height=0, size=0, exifdata=None,
				albumID=album)
		except:
			flash('Error creating photo DB record.')
			return redirect(url('/album/%d' % album))

		id = photo.id

		try:
			data = image.file.read()
			filename = ('%d.jpg' % id)
			fullpath = os.path.join(self.photo_dir, filename)

			fp = open(fullpath, 'wb')
			fp.write(data)
			fp.close()
		except:
			#log.debug(traceback.format_exc())
			flash('Error saving raw image file.')
			if os.path.exists(fullpath):
				os.remove(fullpath)
			photo.destroySelf()
			return redirect(url('/album/%d' % album))

		try:
			pil = IMG.open(StringIO(data))
			pil = self._orient(pil)
			photo.width = pil.size[0]
			photo.height = pil.size[1]
			photo.exifdata = self._exifdate(pil)
		except:
			#log.debug(traceback.format_exc())
			flash('Error opening image, aborting.')
			if os.path.exists(fullpath):
				os.remove(fullpath)
			photo.destroySelf()
			return redirect(url('/album/%d' % album))


		photo.filename = filename
		photo.size = len(data)
		flash('Photo %d uploaded.' % photo.id)

		idx = a.photos().index(photo)
		page = (idx/8)+1

		raise redirect(url('/album/%d?page=%d' % (album, page)))

	@expose()
	@identity.require(identity.in_group('admin'))
	@validate(validators=dict(id=v.Int(not_empty=True)))
	def delete(self, id, tg_errors=None):
		if tg_errors:
			flash('Error: ID is not an int.')
			return redirect(cherrypy.request.headers.get(\
							"Referer", url("/")))

		try:
			photo = Photo.get(id)
		except:
			flash('Error: Photo %d does not exist.')
			return redirect(cherrypy.request.headers.get(\
							"Referer", url("/")))

		albumid = photo.albumID

		fullname = os.path.join(self.photo_dir, photo.filename)
		if os.path.exists(fullname):
			os.remove(fullname)


		album = Album.get(albumid)
		idx = album.photos().index(photo)

		photo.destroySelf()

		page = (idx/8) + 1

		raise redirect(url('/album/%d?page=%d' % (albumid, page)))
	
	def _exifdate(self, img):
		exif = img._getexif()
		o = None
		if exif and exif.has_key(36867):
			o = exif[36867]
		return o

	def _orient(self, img):
		exif = img._getexif()
		o = None
	        if exif and exif.has_key(274):
			o = exif[274]
		
		if o:
			if o == 2 and img.size[0] < img.size[1]:
				img = img.transpose(IMG.FLIP_LEFT_RIGHT)
                        if o == 3 and img.size[0] < img.size[1]:
                                img = img.transpose(IMG.ROTATE_180)
			if o == 4 and img.size[0] < img.size[1]:
				img = img.transpose(IMG.FLIP_TOP_BOTTOM)
			if o == 5 and img.size[0] < img.size[1]:
				img = img.transpose(IMG.ROTATE_270)
				img = img.transpose(IMG.FLIP_LEFT_RIGHT)
                        if o == 6 and img.size[0] > img.size[1]:
                                img = img.transpose(IMG.ROTATE_270)
                        if o == 7 and img.size[0] > img.size[1]:
                                img = img.transpose(IMG.ROTATE_90)
				img = img.transpose(IMG.FLIP_LEFT_RIGHT)
			if o == 8 and img.size[0] > img.size[1]:
				img = img.transpose(IMG.ROTATE_90)		
		return img

	@expose(content_type='image/jpeg')
	@validate(validators=dict(id = v.Int(not_empty=True)))
	def thumb(self, id, tg_errors=None):
		if tg_errors:
			flash('ID is not an integer.')
			raise redirect(request.headers.get('Referer', url('/')))

		try:
			photo = Photo.get(id)
		except:
			flash('Error: Photo ID %d does not exist.' % id)
			id = 0

		if id == 0:
			thumb = IMG.open( \
				os.path.join(self.photo_dir,'nophoto.jpg'))
			thumb.thumbnail((160,100))
			return thumb.tostring('jpeg','RGB')

		if photo.thumbnail == None:
			fullname = os.path.join(self.photo_dir, photo.filename)
			thumb = IMG.open(fullname)
			thumb = self._orient(thumb)
			if thumb.size[0] > thumb.size[1]:
				thumb.thumbnail((160, 100))
			else:
				thumb.thumbnail((100, 160))
			photo.thumbnail = thumb.tostring('jpeg','RGB')
		return photo.thumbnail

	@expose(content_type='image/jpeg')
	@validate(validators = dict(
			width=v.All(
				v.Int(not_empty=True),
				v.OneOf(['1600', '1280', '1024', '800', '640', '426', '320']),
			),
			id = v.Int(not_empty=True)))
	def size(self, width, id, tg_errors=None):
		if tg_errors:
			if 'width' in tg_errors.keys():
				idx = tg_errors.keys().index('width')
				flash('Error: Invalid photo size: %s' % \
						tg_errors.values()[idx])
			else:
				flash('Error: Invalid photo ID.')
			raise redirect(request.headers.get('Referer',url('/')))

		try:
			photo = Photo.get(id)
		except:
			#log.debug(traceback.format_exc())
			flash('Error: Photo %d does not exist.' % id)
			raise redirect(request.headers.get('Referer','/'))

		try:
			fullpath = os.path.join(self.photo_dir, photo.filename)
			img = IMG.open(fullpath)
			if img.size[0] > img.size[1]:
				height = int(img.size[1]* \
					(float(float(width)/float(img.size[0]))))
			else:
				height = width
				width = int(img.size[0]* \
					(float(float(height)/float(img.size[1]))))

			img = self._orient(img)
			out = img.resize((width, height), IMG.ANTIALIAS)
			return out.tostring('jpeg','RGB')
		except:
			#log.debug(traceback.format_exc())
			flash('Error resizing photo, sending full photo.')
			raise redirect(url('/static/photos/%s'% photo.filename))

	@expose(template='.templates.photo')
	@validate(validators=dict(id=v.Int(not_empty=True)))
	def page(self, id, tg_errors=None):
		if tg_errors:
			flash('ID is not an integer.')
			raise redirect(request.headers.get('Referer', url('/')))
		try:
			photo = Photo.get(id)
		except:
			#log.debug(traceback.format_exc())
			flash('Error: Photo %d does not exist.' % id)
			raise redirect(request.headers.get('Referer',url('/')))

		album = Album.get(photo.albumID)
		photos = list(Photo.selectBy(albumID=album.id))
		pids = None
		prev = album.get_prev(photo.id)
		next = album.get_next(photo.id)
		comments = Comment.selectBy(photoID=id)
		cfields = AddCommentFields()
		cfields.append(HiddenField(name='photo', default=str(id))) 
		commentform = TableForm(fields=cfields, \
				action=url('/image/postcomment'))
		ep_values = dict(name=photo.name, desc=photo.descr, \
				album=str(photo.albumID), id=str(photo.id))
		ep_text = 'Save'

		page = (album.photos().index(photo)/8) + 1

		return dict(prev=prev, next=next, photo=photo, album=album,
				comments=comments, commentform=commentform, 
				formtext='Add Comment', 
				editphoto=edit_photo_form,
				editphoto_values=ep_values, 
				editphoto_text=ep_text,
				page = page)

	@expose()
	@validate(validators=dict(
			comment=v.UnicodeString(max=20480),
			photo=v.Int(not_empty=True)))
	def postcomment(self, comment, photo, tg_errors=None):
		if tg_errors:
			if 'comment' in tg_errors.keys():
				flash('Error: Comment was too long (>20480 chars).')
				raise redirct(request.headers.get('Referer','/'))
			else:
				flash('Error: Photo %d does not exist' % photo)
				raise redirct(request.headers.get('Referer','/'))

		if not identity.current.anonymous:
			userid = int(identity.current.user.id)
		else:
			userid = 1
		try:
			Comment(comment=comment, date=datetime.now(),\
				photoID=photo, \
				userID=userid)
		except:
			flash('Error: Comment not saved.')
			raise redirct(request.headers.get('Referer','/'))

		flash('Comment Saved.')
		return redirect(url('/image/page/%d' % photo))

	@expose(content_type='image/jpeg')
	@validate(validators=dict(id = v.Int(not_empty=True)))
	def default(self, id, tg_errors=None):
		if tg_errors:
			flash('Error: Photo does not exist.')
			raise redirect(request.headers.get('Referer','/'))
		return redirect(url('/static/photos/%d.jpg' % id))


class Root(controllers.RootController):

    album=ShowAlbum()
    image=Image()

    @expose(template=".templates.index")
    @validate(validators=dict(page=v.Int()))
    def index(self, page='1', tg_errors = None):
	allalbums=list(Album.select(orderBy="-date"))
	page, pages, start, end = paginate(page, len(allalbums))
	albums=allalbums[start:end]
        return dict(albums=albums, page=page, pages=pages, \
			form=add_album_form, formtext='Create')

    @expose(content_type="text/plain")
    def albums(self):
	albums=Album.select(orderBy="id")
	page = ''
	for a in albums:
		page += '%d,%s\n' % (a.id, a.name)
	return page

    @expose(template=".templates.login")
    def login(self, forward_url=None, previous_url=None, *args, **kw):

        if not identity.current.anonymous \
            and identity.was_login_attempted() \
            and not identity.get_identity_errors():
            raise redirect(forward_url)

        forward_url=None
        previous_url= request.path

        if identity.was_login_attempted():
            msg=_("The credentials you supplied were not correct or "
                   "did not grant access to this resource.")
        elif identity.get_identity_errors():
            msg=_("You must provide your credentials before accessing "
                   "this resource.")
        else:
            msg=_("Please log in.")
            forward_url= request.headers.get("Referer", "/")
            
        response.status=403
        return dict(message=msg, previous_url=previous_url, logging_in=True,
                    original_parameters=request.params,
                    forward_url=forward_url)

    @expose()
    def logout(self):
        identity.current.logout()
        raise redirect(url("/"))
