from datetime import datetime
from turbogears.database import PackageHub
from sqlobject import *
from turbogears import identity 
import random

hub = PackageHub("shutterbug")
__connection__ = hub

class Album(SQLObject):
	class sqlmeta:
		table = "albums"
	name = UnicodeCol()
	descr = UnicodeCol()
	date = DateTimeCol(default = datetime.now)

	def photos(self, shuffle=False):
		if shuffle:
			p=Photo.select(Photo.q.albumID==self.id,orderBy='id')
			lp = list(p)
			random.shuffle(lp)
			return lp

		return list(Photo.select(Photo.q.albumID==self.id,orderBy='id'))

	def size(self):
		return len(self.photos())

	def photo_index(self, pid):
		photo = Photo.get(pid)
		return (self.photos().index(photo))+1

	def get_prev(self, pid):
		photo = Photo.get(pid)
		photos = self.photos()
		if photo in photos:
			idx = photos.index(photo)
			if idx:
				return photos[idx-1].id
		return 0

	def get_next(self, pid):
		photo = Photo.get(pid)
		photos = self.photos()
		if photo in photos:
			idx = photos.index(photo)
			if idx < (len(photos)-1):
				return photos[idx+1].id
		return 0

			

class Photo(SQLObject):
	class sqlmeta:
		table = "photos"
	name = UnicodeCol()
	descr = UnicodeCol()
	date = DateTimeCol(default = datetime.now)
	width = IntCol(default=0)
	height = IntCol(default=0)
	size = IntCol(default=0)
	filename = UnicodeCol()
	exifdata = BLOBCol(default=None)
	thumbnail = BLOBCol(default=None)
	album = ForeignKey('Album', cascade=True)
	

class Comment(SQLObject):
	class sqlmeta:
		table = "comments"
	comment = UnicodeCol(length=10240)
	date = DateTimeCol(default=datetime.now)
	user = ForeignKey('User', cascade=None)
	photo = ForeignKey('Photo', cascade=True)

# identity models.
class Visit(SQLObject):
    class sqlmeta:
        table = "visit"

    visit_key = StringCol(length=40, alternateID=True,
                          alternateMethodName="by_visit_key")
    created = DateTimeCol(default=datetime.now)
    expiry = DateTimeCol()

    def lookup_visit(cls, visit_key):
        try:
            return cls.by_visit_key(visit_key)
        except SQLObjectNotFound:
            return None
    lookup_visit = classmethod(lookup_visit)

class VisitIdentity(SQLObject):
    visit_key = StringCol(length=40, alternateID=True,
                          alternateMethodName="by_visit_key")
    user_id = IntCol()


class Group(SQLObject):
    """
    An ultra-simple group definition.
    """

    # names like "Group", "Order" and "User" are reserved words in SQL
    # so we set the name to something safe for SQL
    class sqlmeta:
        table = "tg_group"

    group_name = UnicodeCol(length=16, alternateID=True,
                            alternateMethodName="by_group_name")
    display_name = UnicodeCol(length=255)
    created = DateTimeCol(default=datetime.now)

    # collection of all users belonging to this group
    users = RelatedJoin("User", intermediateTable="user_group",
                        joinColumn="group_id", otherColumn="user_id")

    # collection of all permissions for this group
    permissions = RelatedJoin("Permission", joinColumn="group_id", 
                              intermediateTable="group_permission",
                              otherColumn="permission_id")


class User(SQLObject):
    """
    Reasonably basic User definition. Probably would want additional attributes.
    """
    # names like "Group", "Order" and "User" are reserved words in SQL
    # so we set the name to something safe for SQL
    class sqlmeta:
        table = "tg_user"

    user_name = UnicodeCol(length=16, alternateID=True,
                           alternateMethodName="by_user_name")
    email_address = UnicodeCol(length=255, alternateID=True,
                               alternateMethodName="by_email_address")
    display_name = UnicodeCol(length=255)
    password = UnicodeCol(length=40)
    created = DateTimeCol(default=datetime.now)

    # groups this user belongs to
    groups = RelatedJoin("Group", intermediateTable="user_group",
                         joinColumn="user_id", otherColumn="group_id")

    def _get_permissions(self):
        perms = set()
        for g in self.groups:
            perms = perms | set(g.permissions)
        return perms

    def _set_password(self, cleartext_password):
        "Runs cleartext_password through the hash algorithm before saving."
        hash = identity.encrypt_password(cleartext_password)
        self._SO_set_password(hash)

    def set_password_raw(self, password):
        "Saves the password as-is to the database."
        self._SO_set_password(password)



class Permission(SQLObject):
    permission_name = UnicodeCol(length=16, alternateID=True,
                                 alternateMethodName="by_permission_name")
    description = UnicodeCol(length=255)

    groups = RelatedJoin("Group",
                        intermediateTable="group_permission",
                         joinColumn="permission_id", 
                         otherColumn="group_id")


